"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
Object.defineProperty(exports, "__esModule", { value: true });
const fastcar_core_1 = require("fastcar-core");
const utils_1 = require("fastcar-core/utils");
const annotation_1 = require("fastcar-core/annotation");
const fastcar_server_1 = require("fastcar-server");
const SocketFactory_1 = require("./SocketFactory");
const SocketEnum_1 = require("../../constant/SocketEnum");
const nanoid_1 = require("nanoid");
const SocketMsgStatus_1 = require("../../constant/SocketMsgStatus");
const RpcConfig_1 = require("../../types/RpcConfig");
/***
 * @version 1.0 用于集成各个不同类型的socket和实现丰富的消息逻辑表达
 */
class SocketManager {
    constructor() {
        this.serverMap = new Map();
        this.clientSessionMap = new Map();
        this.channels = new Map();
    }
    async auth(username, password, session) {
        if (!this.msgCallBack) {
            return false;
        }
        return await this.msgCallBack.auth(username, password, session);
    }
    bind(msgCallBack) {
        this.msgCallBack = msgCallBack;
    }
    getLogger() {
        return this.logger;
    }
    connect(sessionId) {
        let session = this.getClientSession(sessionId);
        if (session) {
            this.msgCallBack.connect(session);
        }
    }
    disconnect(sessionId, reason, force) {
        let session = this.clientSessionMap.get(sessionId);
        if (session) {
            let serverId = session.serverId;
            if (force) {
                //通知server管理层下线
                let socketServer = this.serverMap.get(serverId);
                if (socketServer) {
                    socketServer.kickConnect({ sessionId, reason });
                }
            }
            else {
                this.msgCallBack.disconnect(session, reason);
            }
            this.clientSessionMap.delete(sessionId);
        }
    }
    handleMsg(sessionId, msg) {
        let session = this.clientSessionMap.get(sessionId);
        if (session) {
            this.msgCallBack.handleMsg(session, msg);
        }
    }
    getClientSession(sessionId) {
        return this.clientSessionMap.get(sessionId) || null;
    }
    getSocketServer(serverId) {
        return this.serverMap.get(serverId) || null;
    }
    getSocketServerBySessionId(sessionId) {
        let session = this.clientSessionMap.get(sessionId);
        if (!session) {
            return null;
        }
        let server = this.serverMap.get(session.serverId);
        return server || null;
    }
    getSocketClientBySessionId(sessionId) {
        let server = this.getSocketServerBySessionId(sessionId);
        if (server) {
            return server.getSession(sessionId) || null;
        }
        return null;
    }
    //会话是否在线
    sessionOnline(sessionId) {
        let session = this.clientSessionMap.get(sessionId);
        if (!session) {
            return false;
        }
        return this.serverMap.has(session.serverId);
    }
    //发送消息
    async sendMsg(sessionId, msg = { mode: RpcConfig_1.InteractiveMode.notify, url: "" }) {
        let server = this.getSocketServerBySessionId(sessionId);
        if (server) {
            let flag = await server.sendMsgBySessionId(sessionId, msg);
            if (flag) {
                return SocketMsgStatus_1.SocketMsgStatus.success;
            }
            return SocketMsgStatus_1.SocketMsgStatus.fail;
        }
        return SocketMsgStatus_1.SocketMsgStatus.offline;
    }
    //加入频道
    joinChannel(sessionId, channel) {
        let session = this.getClientSession(sessionId);
        if (!session) {
            return false;
        }
        if (!this.channels.has(channel)) {
            this.channels.set(channel, []);
        }
        let sessionIds = this.channels.get(channel);
        if (!sessionIds?.includes(sessionId)) {
            sessionIds?.push(sessionId);
        }
        return true;
    }
    leaveChannel(sessionId, channel) {
        let session = this.getClientSession(sessionId);
        if (!session) {
            return;
        }
        let sessionIds = this.channels.get(channel);
        if (!sessionIds) {
            return;
        }
        let index = sessionIds.indexOf(sessionId);
        if (index != -1) {
            sessionIds.splice(index, 1);
            if (sessionIds.length == 0) {
                this.channels.delete(channel);
            }
        }
    }
    //根据sessionId获取所有的渠道
    getChannelBySessionId(sessionId) {
        if (!this.clientSessionMap.has(sessionId)) {
            return [];
        }
        let list = [];
        this.channels.forEach((sessionIds, channel) => {
            if (sessionIds.includes(sessionId)) {
                list.push(channel);
            }
        });
        return list;
    }
    sendMsgByChannel(channel, msg, excludeIds = []) {
        let sessionIds = this.channels.get(channel);
        if (!sessionIds || sessionIds.length == 0) {
            return;
        }
        sessionIds.forEach((sessionId) => {
            if (excludeIds.includes(sessionId)) {
                return;
            }
            this.sendMsg(sessionId, msg);
        });
    }
    getSocketServerConfig(serverId) {
        return this.serverMap.get(serverId)?.getConfig();
    }
    checkSocketConfig(socketConfig) {
        if (!Array.isArray(socketConfig)) {
            return false;
        }
        let socketValue = [];
        Reflect.ownKeys(SocketEnum_1.SocketEnum).forEach((k) => {
            socketValue.push(Reflect.get(SocketEnum_1.SocketEnum, k));
        });
        let flag = socketConfig.every((item) => {
            if (!socketValue.includes(item.type)) {
                return false;
            }
            if (!utils_1.ValidationUtil.isString(item.id)) {
                return false;
            }
            if (!item.server) {
                item.server = {};
            }
            if (!item.extra) {
                item.extra = {};
            }
            item.server = Object.assign({ port: 80, protocol: fastcar_server_1.Protocol.http }, item.server);
            return true;
        });
        if (!flag) {
            return false;
        }
        return true;
    }
    async start(socketConfig) {
        //读取配置进行socket初始化
        if (!this.checkSocketConfig(socketConfig)) {
            let err = new Error("socket config is error");
            this.logger.error(err);
            throw err;
        }
        socketConfig.forEach(async (item) => {
            let SocketClass = (0, SocketFactory_1.SocketServerFactory)(item.type);
            if (SocketClass) {
                let server = new SocketClass(item, this);
                await server.listen();
                this.logger.info(`${item.type} server [${item.id}] is running in ${item.server.port}`);
                this.serverMap.set(item.id, server);
            }
            else {
                this.logger.error(`This type [${item.type}] of connection is not supported`);
            }
        });
    }
    async stop() {
        for (let [serverId, server] of this.serverMap) {
            await server.close();
            this.logger.info(`${serverId} is close`);
        }
        this.clientSessionMap.clear();
        this.serverMap.clear();
    }
    //创建一个初始的会话
    createSession(serverId) {
        let id = (0, nanoid_1.nanoid)();
        let session = {
            sessionId: id,
            serverId,
            connectedTime: Date.now(),
            settings: new Map(),
        };
        this.clientSessionMap.set(id, session);
        return session;
    }
    //销毁会话
    deleteSession(sessionId) {
        this.clientSessionMap.delete(sessionId);
    }
    //获取一个网络服务器
    createNetServer(config, cb) {
        let server = this.netServer.createServer(config, cb);
        if (!server) {
            let err = new Error(`create net server fail by ${JSON.stringify(config)}`);
            this.logger.error(err);
            throw err;
        }
        return server;
    }
}
__decorate([
    (0, annotation_1.Log)("socket"),
    __metadata("design:type", fastcar_core_1.Logger)
], SocketManager.prototype, "logger", void 0);
__decorate([
    annotation_1.CallDependency,
    __metadata("design:type", fastcar_server_1.ServerApplication)
], SocketManager.prototype, "netServer", void 0);
exports.default = SocketManager;
